/*
** Module   : Variable.java
** Abstract : Implementation of a user variable accessible from expressions
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050201   @19690 A helper class to manage a variable during
**                           expression compilation.
** 002 ECF 20050509   @21354 Complete rewrite. Implements a user variable
**                           of arbitrary type. Can be set to read/write
**                           or read-only. Can optionally assign an
**                           arbitrary expression to initialize its value.
** 003 ECF 20050604   @21410 Removed construction-time type verification
**                           between declared type and initializer return
**                           type. This is left to the setValue method,
**                           which makes us more flexible in accepting
**                           initializers which return a wider type than
**                           the variable's declared type. Since the
**                           initializer is run immediately from the
**                           constructor, this effectively does not change
**                           the timing of data type verification.
** 004 GES 20130203          Modified to utilize a new Initializer class. This
**                           allows a hard coded instance to be set as the value
**                           in addition to the current expression evaluation
**                           mechanism.
*/

package com.goldencode.expr;

/**
 * Implementation of a user variable which is accessible from expressions.
 * A variable has an immutable name. It can be of any non-primitive data
 * type. From within an expression, it may be readable and writable, or
 * read-only. Regardless of mutability from within expressions, the value
 * of a variable may always be changed {@link #setValue programmatically}.
 * <p>
 * A variable may have an optional initializer associated with it. This can
 * be a simple instance that is assigned at construction and when reset. Or
 * can be an expression which is evaluated to generate the initial value. 
 * Any instance or expression will be assigned each time the {@link #reset}
 * method is called. The result of executing the initializer expression will
 * be stored as the new value of the variable. The variable is reset
 * automatically at construction. If no initializer expression is provided,
 * resetting the variable will set its value to <code>null</code>. The
 * initializer expression for a variable may be arbitrarily complex, but
 * it may not refer to its containing variable, and it may only refer to
 * libraries and variables which previously have been registered with the
 * symbol resolver.
 * <p>
 * Variables are stored within variable pools inside a {@link SymbolResolver
 * symbol resolver}. A variable is added to a variable pool through the
 * {@link SymbolResolver#registerVariable} method. It is that method which
 * instantiates new <code>Variable</code> instances. Although a data type
 * is optional during registration if an initializer expression is provided,
 * it is recommended to provide a data type if it is possible that a more
 * specific data type is needed than may be returned by the initializer.
 * Otherwise, methods available only in the more specific type will not be
 * accessible to expressions which reference this variable.
 */
public class Variable
{
   /** Symbol by which this variable will be referenced in user expressions. */
   private String name = null;
   
   /** Data type. */
   private Class type = null;
   
   /** Current value. */
   private Object value = null;
   
   /** Whether the variable is read-only from within expressions. */
   private boolean readOnly = false;
   
   /** Used to initialize and reset current value. */
   private Initializer initializer = null;
   
   /**
    * Construct a user variable with the given name and type, and initialize
    * its value using the supplied expression. If both a data type and an
    * initializer expression are provided, the return type of the expression
    * must be {@link Function#isAssignmentCompatible assignment compatible}
    * with the declared data type. If the data type is omitted, the return
    * type of the initializer will be used as the variable's data type.
    *
    * @param   name
    *          Case-sensitive name by which this variable will be referenced
    *          within user expressions.
    * @param   type
    *          Data type of this variable. Required if
    *          <code>initializer</code> is <code>null</code>.
    * @param   initializer
    *          The expression or value used to initialize and reset the value
    *          of this variable. May be <code>null</code>, in which case the
    *          starting value of the variable is <code>null</code>, and the
    *          value to which the variable will be {@link #reset reset} is also
    *          <code>null</code>. Required if <code>type</code> is
    *          <code>null</code>.
    *
    * @throws  IllegalArgumentException
    *          if <code>type</code> represents a primitive class; if
    *          <code>type</code> was <code>null</code> and could not be
    *          determined from <code>initializer</code>'s return type; if
    *          <code>type</code> was not <code>null</code> and was
    *          incompatible with <code>initializer</code>'s return type.
    * @throws  UnresolvedSymbolException
    *          if initializer expression cannot resolve a mandatory symbol.
    * @throws  ClassCastException
    *          if initializer expression produces a non-<code>null</code>
    *          result which is not appropriate to the data type of this
    *          variable.
    */
   Variable(String name, Class type, Initializer initializer, boolean readOnly)
   {
      if (type != null && type.isPrimitive())
      {
         throw new IllegalArgumentException(
            "User variable (" + name +
            ") cannot contain a primitive value (" + type + ")");
      }
      
      // If type is not declared, use initializer's return type as the type.
      if (initializer != null)
      {
         Class returnType = initializer.getReturnType();
         
         if (type == null)
         {
            type = returnType;
         }
      }
      
      if (type == null)
      {
         throw new IllegalArgumentException(
            "Data type for variable '" + name +
            "' was not provided or could not be determined from initializer");
      }
      
      this.name = name;
      this.type = type;
      this.initializer = initializer;
      this.readOnly = readOnly;
      
      // Set the value of the variable using the initializer. If no
      // initializer is defined, set it to null.
      reset();
   }
   
   /**
    * Get the name of this variable. This is the symbol by which the variable
    * is referenced in expressions.
    *
    * @return  Variable name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Get the data type of this variable.
    *
    * @return  Data type.
    */
   public Class getType()
   {
      return type;
   }
   
   /**
    * Get the current value of this variable.
    *
    * @return  Current value.
    */
   public Object getValue()
   {
      return value;
   }
   
   /**
    * Indicate whether this variable is read-only from within expressions.
    * Note that the value of a read-only variable can <b>always</b> be
    * changed programatically through the various methods provided by this
    * class.  Read-only in this context refers only to whether a value may
    * be assigned to this variable from within an expression.
    *
    * @return  <code>true</code> if this variable is read-only, else
    *          <code>false</code>.
    */
   public boolean isReadOnly()
   {
      return readOnly;
   }
   
   /**
    * Set the current value of this variable.
    *
    * @param   value
    *          New value. Must be assignment-compatible with this variable's
    *          data type.
    *
    * @throws  ClassCastException
    *          if type of <code>value</code> is not compatible with this
    *          variable's data type.
    */
   public void setValue(Object value)
   {
      if (value != null && !type.isInstance(value))
      {
         throw new ClassCastException(
            "Cannot set value for variable '" + name + "' (" +
            type.getName() + ") to type " + value.getClass().getName());
      }
      
      this.value = value;
   }
   
   /**
    * Reset the current value of this variable to its default, using the
    * initializer expression provided at construction. If no initializer
    * was provided, the current value becomes <code>null</code>.
    *
    * @throws  UnresolvedSymbolException
    *          if initializer expression cannot resolve a mandatory symbol.
    * @throws  ClassCastException
    *          if initializer expression produces a non-<code>null</code>
    *          result which is not appropriate to the data type of this
    *          variable.
    */
   public void reset()
   {
      setValue(initializer == null ? null : initializer.resolve());
   }
   
   /**
    * Get a string representation of the value contained within this
    * variable.  This method is delegated to the <code>toString</code>
    * implementation of the value object.  A <code>null</code> value is
    * safely handled.
    *
    * @return  String representation of the value contained within this
    *          variable.
    */
   public String toString()
   {
      return String.valueOf(value);
   }
}
